class ProjectDefaults:
    PROJECT_NAME = "windmill-project"
    PROJECT_CONF = "windmill.conf"
    WML_FOLDER = "wmls"
    DAGS_FOLDER = "dags"
    OPERATORS_FOLDER = "custom_operators"


class ServerDefaults:
    HOST_ADDRESS = "localhost"
    HOST_PORT = 8000
    PROJECT_CONF = ProjectDefaults.PROJECT_CONF



import os
from collections import namedtuple

import yaml
from marshmallow import fields, Schema

from .base_config import BaseConfig
from ..constants import ProjectDefaults
from ..exceptions import InitError


class ProjectConfig(BaseConfig):
    schema = Schema.from_dict(
        {
            # CLI Arguments
            "name": fields.Str(
                missing=ProjectDefaults.PROJECT_NAME,
                help="Name of the root project folder",
            ),
            "wml_dir": fields.Str(
                missing=ProjectDefaults.WML_FOLDER,
                help="Folder to store windmill WML files",
            ),
            "dags_dir": fields.Str(
                missing=ProjectDefaults.DAGS_FOLDER,
                help="Folder to store generated YML DAG files",
            ),
            "operators_dir": fields.Str(
                missing=ProjectDefaults.OPERATORS_FOLDER,
                help="Folder to store custom operator files",
            ),
        }
    )

    def __init__(
        self,
        name: str,
        wml_dir: str,
        dags_dir: str,
        operators_dir: str,
        conf_file: str = ProjectDefaults.PROJECT_CONF,
        *args,
        **kwargs,
    ):
        """Handler for project file
        
        Args:
            ...Project.schema
            conf_file (str, optional): Default project config filename. Defaults to ProjectDefaults.PROJECT_CONF.
        """
        self.name = name
        self.wml_dir = wml_dir
        self.dags_dir = dags_dir
        self.operators_dir = operators_dir
        self.conf_file = conf_file

    @staticmethod
    def from_conf_file(filename):
        try:
            with open(filename, "r+") as f:
                return ProjectConfig(**yaml.load(f))
        except Exception as e:
            raise InitError("This directory is not a valid windmill project") from e



import inspect
import logging
import re
from copy import deepcopy

from docstring_parser import parse

from ..exceptions import DocstringParseError
from ..models.schemas.app_schemas import (
    validate_parameter_type,
    VALID_PARAMETER_TYPES,
    PARAMETER_MAPPINGS,
)


class ClassParser:
    @classmethod
    def fix_types(cls, typ: str):
        if typ.lower().startswith("a "):
            typ = typ[2:]

        for root, mapping in PARAMETER_MAPPINGS.items():
            if typ.startswith(root):
                typ = mapping

        if typ in PARAMETER_MAPPINGS:
            typ = PARAMETER_MAPPINGS[typ]

        for root in VALID_PARAMETER_TYPES:
            if typ.startswith(root):
                typ = root
        return typ

    @classmethod
    def fix_docstring(cls, docs: str):
        """Helper to convert Airflow __doc__ into valid docstring_parser
        docstrings 

        Note that this will mung the parsed datatypes, defaulting to a str data
        type.

        Args:
            docs (str): Valid Airflow docstring, eg
                Some description...

                :param param1: some param
                :type param1: e.g. str
                ...

        return str: Reformatted str eg
                Some description...

                :param str param1: some param
                :type param1: e.g. str
                ...
        """
        type_re = r":type (.*): (.*)"

        for param, typ in re.findall(type_re, docs):
            typ = cls.fix_types(typ)
            if not validate_parameter_type(typ) or not typ:
                logging.warning(f"Unable to parse field {typ}")
                typ = "str"
            docs = docs.replace(f":param {param}:", f":param {typ} {param}:")
        return docs

    @staticmethod
    def _parse_class_docstring(cls, inherited):
        doc_string = cls.__doc__
        if not doc_string:
            raise DocstringParseError(f"Class {cls} has no docstring")

        if cls.__module__ == "airflow.operators.pig_operator":
            print("hello")

        if doc_string:
            fixed_docs = ClassParser.fix_docstring(doc_string)
            docs = parse(fixed_docs)
            return [
                cls.__name__,
                {
                    "module": cls.__module__,
                    "description": "\n".join(
                        (docs.short_description or "", docs.long_description or "")
                    ),
                    "parameters": [
                        {
                            "id": p.arg_name,
                            "type": p.type_name or "str",
                            "description": p.description,
                            "inheritedFrom": cls.__name__ if inherited else None,
                        }
                        for p in docs.params
                    ],
                },
            ]
        else:
            raise DocstringParseError(f"Unable to parse docstring for class {cls}")

    @staticmethod
    def parse_class(cls, inherit_until=None, inherited=False):
        """Parses class docstrings and parent classes to get class properties
        This will add default values from the func definition as well
        
        Args:
            cls (Class): The class to parse
            inherit_until (Class, optional): If provided will scrape parent class properties, 
                                             until inherit_until class is reached (e.g. BaseOperator)
            inherited (bool, optional): Parameters will include the inheritedFrom property
        
        Returns:
            list: [class name, properties dict]
        """
        name, props = ClassParser._parse_class_docstring(cls, inherited)

        # Get default parameter values from class
        class_sig = inspect.signature(cls)
        for docstring_param in props["parameters"]:
            param = class_sig.parameters.get(docstring_param["id"], None)
            if param:
                if param.default == inspect._empty:
                    docstring_param["required"] = True
                else:
                    docstring_param["default"] = param.default

        if inherit_until:
            additional_params = []
            for parent in [c for c in cls.__bases__ if c == inherit_until]:
                _, parent_props = ClassParser.parse_class(parent, inherit_until, True)
                for param in parent_props["parameters"]:
                    if param["id"] not in [
                        p["id"] for p in (additional_params + props["parameters"])
                    ]:
                        additional_params.append(param)
            props["parameters"].extend(additional_params)

        return name, props



from abc import ABC, abstractstaticmethod

from marshmallow import Schema


class BaseConfig(ABC):
    """Helper class to manage CLI parsing and serialisation 
    """

    schema: Schema

    def dump(self):
        return self.schema().dump(self)

    @classmethod
    def load(cls, *args, **kwargs):
        """Runs Schema.load method, but converts the dictionary result
        into an object with named fields.
        
        Returns:
            ProjectObj: Instantiated object with the same fields as the parent schema
        """
        # Use schema to validate/translate
        data = cls.schema().load(kwargs, unknown="IGNORE")

        # Pass on additional params that aren't in the schema
        # This would include args that are not intended for the cli
        full_params = {**data, **{k: v for k, v in kwargs.items() if k not in data}}
        return cls(**full_params)

    @classmethod
    def to_cli_args(cls):
        """Converts this config object into a DeCli arguments list. If a field
        has `ignore_cli=True` metadata then it will not be included. If not specified
        with arg_type the default type is assumed as string.

        TODO: Implicit field typings (e.g. fields.Str -> str())
        
        Returns:
            List<dict>: List of command parameters 
        """
        return [
            {
                "name": f"--{f_name.replace('_', '-')}",
                "help": f_spec.metadata.get("help", ""),
                "default": f_spec.missing,
                "type": f_spec.metadata.get("arg_type", str),
            }
            for f_name, f_spec in cls.schema().fields.items()
        ]



import json
import logging
import os

from flask import Flask, jsonify, make_response, request, send_from_directory
from flask_cors import CORS
from marshmallow import EXCLUDE
from marshmallow.exceptions import ValidationError

from ...config.project_config import ProjectConfig
from ...exceptions import DagHandlerValidationError
from ...models.dags.dag_handler import DagHandler
from ...models.operators.operator_index import OperatorIndex
from ...models.schemas.app_schemas import OperatorSchema, MinimalWmlSchema


_operator_index: OperatorIndex = None


def get_operator_index() -> OperatorIndex:
    global _operator_index

    if not _operator_index:
        _operator_index = OperatorIndex()
    return _operator_index


app = Flask(__name__, static_folder="../app/dist/")

#######################################################################
# Serve React App
#######################################################################
@app.route("/")
def serve():
    return send_from_directory(app.static_folder, "index.html")


#######################################################################
# API Endpoints
#######################################################################
@app.route("/v1/operators", methods=["GET"])
def get_operators():
    """Retrieves a JSON list of all available Airflow Operators
    
    Returns:
        List(OperatorIndex)
    """
    logging.info(f"GET /v1/operators")

    return jsonify(get_operator_index().marshall_operator_list()), 200


@app.route("/v1/wml/", methods=["GET"])
@app.route("/v1/wml/<name>", methods=["GET"])
def get_wmls(name=None):
    logging.info(f"GET /v1/wml/{name}")

    if name:
        f_path = os.path.join(app.config["project_conf"].wml_dir, name)
        if os.path.exists(f_path):
            with open(f_path, "r") as f:
                data = json.load(f)
            return jsonify(data), 200
        else:
            return f"File {f_path} not found", 404
    else:
        return jsonify(os.listdir(app.config["project_conf"].wml_dir)), 200


@app.route("/v1/wml/<name>", methods=["POST"])
def post_wml(name):
    logging.info(f"POST /v1/wml/{name}")

    content = request.json
    # TODO: Is validation required here or just on DAG?

    f_path = os.path.join(app.config["project_conf"].wml_dir, name)
    with open(f_path, "w") as f:
        json.dump(content, f)

    return "Created", 201


@app.route("/v1/dag", methods=["GET"])
def get_dagspec():
    logging.info(f"GET /v1/dag/")

    return jsonify(DagHandler.marshall_dag_docstring()), 200


@app.route("/v1/dag/<name>", methods=["POST"])
def post_dag(name):
    logging.info(f"POST /v1/dag/{name}")

    res = post_wml(name)
    if res[1] == 201:
        f_path = os.path.join(app.config["project_conf"].wml_dir, name)
        with open(f_path, "r") as f:
            wml_dict = json.load(f)

        try:
            wml_dict_parsed = MinimalWmlSchema().load(
                wml_dict, partial=False, unknown=EXCLUDE
            )
        except ValidationError as e:
            logging.exception("Error parsing WML")
            return "Unable to deserialise WML contents", 400

        try:
            dag_handler = DagHandler.load_from_wml(wml_dict_parsed)
            py_content = dag_handler.compile_to_python()
        except DagHandlerValidationError as e:
            logging.exception(f"Unable to convert WML '{name}' to DAG")
            return f"Error: {e}", 400
        except Exception as e:
            logging.exception(f"Unknwon error while converting WML {name}")
            return f"Internal error converting WML", 500

        f_path = os.path.join(
            app.config["project_conf"].dags_dir, f"{dag_handler.snake_name}.py"
        )
        with open(f_path, "w") as f:
            f.write(py_content)

        return "Created", 201
    else:
        return res


#######################################################################
# Build App
#######################################################################
def build_app(proj_conf: ProjectConfig, dev_server=False):
    app.config["project_conf"] = proj_conf

    if dev_server:
        logging.warning("Running a dev-build with CORS enabled")
        CORS(app)

    return app



from marshmallow import Schema, fields

from .base_config import BaseConfig
from .project_config import ProjectConfig
from ..constants import ServerDefaults


class RunConfig(BaseConfig):
    schema = Schema.from_dict(
        {
            "port": fields.Int(missing=ServerDefaults.HOST_PORT, help="Bind port"),
            "hostname": fields.Str(
                missing=ServerDefaults.HOST_ADDRESS, help="Bind Address"
            ),
            "conf_file": fields.Str(
                missing=ServerDefaults.PROJECT_CONF,
                help="Name of config file in this directory",
            ),
        }
    )

    def __init__(
        self,
        port: int,
        hostname: str,
        conf_file: str,
        run_dev_server=False,
        *args,
        **kwargs
    ):
        """Wrapper around run-server
        
        Args:
            ...RunConfig.Schema
            run_dev_server (bool) : If True will run back-end with CORS on Flask (No Gunicorn)
        """
        self.port = port
        self.hostname = hostname
        self.conf_file = conf_file
        self.run_dev_server = run_dev_server

        self.project_conf = ProjectConfig.from_conf_file(self.conf_file)



import logging
import os

import yaml

from ..config.project_config import ProjectConfig
from ..exceptions import InitError


def join(*args):
    return os.path.join(*args)


class CreateProject:
    def __init__(self, proj: ProjectConfig):
        """Creates a new project directory 
        
        Args:
            proj (Project): Project config object
        
        Raises:
            InitError: If project directory cannot be created
        """
        if os.path.exists(proj.name):
            if len(os.listdir(proj.name)) == 0:
                logging.info(f"Project directory  already exists...")
            else:
                raise InitError(f"Non-empty directory '{proj.name}' already exists.")
        else:
            logging.info(f"Creating project directory '{proj.name}'")
            os.mkdir(proj.name)

        logging.info(f"Creating project subfolders...")
        os.mkdir(join(proj.name, proj.wml_dir))
        os.mkdir(join(proj.name, proj.dags_dir))
        os.mkdir(join(proj.name, proj.operators_dir))

        logging.info(f"Creating project config file '{proj.conf_file}'")
        with open(join(proj.name, proj.conf_file), "w+") as f:
            yaml.dump(proj.dump(), f)

        logging.info(f"Succesfully created project '{proj.name}'")



from abc import ABC, abstractproperty
from copy import deepcopy
from typing import List

import black
from airflow.models.dag import DAG
from inflection import underscore
from jinja2 import Environment, PackageLoader, select_autoescape
from marshmallow import Schema, fields
from networkx import DiGraph, is_directed_acyclic_graph, dag_longest_path

from ..schemas.app_schemas import DagSchema, OperatorParameterSchema
from ...exceptions import DagHandlerValidationError
from ...utils.class_parser import ClassParser

_op_schema = OperatorParameterSchema()


class _ParamHandler(ABC):
    def __init__(self, params):
        """Helper class to deal with parameter rendering
        """
        self.params = params

    @abstractproperty
    def snake_name(self):
        raise NotImplementedError("Must be implemented")

    @staticmethod
    def parameter_list_to_dict(params):
        """Validates a list of App Parameters and returns a dict of parameter keys and values
        
        Args:
            params (list): List of Parameters matching OperatorParameterSchema

        Raises:
            DagHandlerValidationError: If a required parameter doesn't have a value
        
        Returns:
            [dict]: Mapping between non-default fields and values
        """
        params = [_op_schema.load(param) for param in params]
        for param in params:
            if param.get("required", False) and param.get("value") == None:
                raise DagHandlerValidationError(
                    f"'{param['id']}' is a required parameter'"
                )
        return {
            param["id"]: param
            for param in params
            if param.get("value") and param["value"] != param.get("default")
        }

    def param_to_callable(self, param_name):
        return f"{self.snake_name}_{underscore(param_name)}_callable"

    @property
    def callables(self):
        return {
            k: {"method": v["value"], "name": self.param_to_callable(k)}
            for k, v in self.params.items()
            if v["type"] == "python_callable"
        }

    @property
    def params_to_kwargs(self):
        """Converts task parameters into a valid kwargs expression. Relies on double asterixes 
        so that we don't need to worry about quotes 
        """
        non_callables = {
            k: v["value"] for k, v in self.params.items() if k not in self.callables
        }
        callables = ", ".join(
            [f"{k} = {self.param_to_callable(k)}" for k in self.callables]
        )
        if callables:
            return f"**{non_callables}, {callables}"
        return f"**{non_callables}"


class TaskHandler(_ParamHandler):
    def __init__(self, node_id, operator_type, module, task_params):
        """Schema for a Task Instance, that can be loaded from an App Node object
        
        Args:
            node_id (str): Internal Node ID used by LinkInstances
            operator_type (str): Name of the operator class
            module (str): Module path to import operator
            task_params (dict): List of Task Parameters
        """
        self.node_id = node_id
        self.operator_type = operator_type
        self.module = module
        self.params = task_params

    @property
    def snake_name(self):
        """snakecase version of task_params["task_id"]
        
        Returns:
            [str]: task_name
        """
        return underscore(self.params["task_id"]["value"])

    @classmethod
    def load_from_node(cls, node_dict):
        """Creates a TaskHandler from a WML Node Dictionary
        
        Args:
            node_dict (dict): Dictionary that matches app_schemas.NodeSchema
        
        Returns:
            [TaskHandler]: The converted TaskHandler object
        """
        node_id = node_dict["id"]
        operator_type = node_dict["type"]
        module = node_dict["properties"]["module"]
        task_params = cls.parameter_list_to_dict(node_dict["properties"]["parameters"])

        if not task_params.get("task_id"):
            raise DagHandlerValidationError(
                f"Node {node_id} missing required parameter task_id"
            )

        return TaskHandler(node_id, operator_type, module, task_params)


class Links:
    def __init__(self, graph):
        """Metadata describing the links in a DAG between tasks
        
        Args:
            link_graph (Digraph): Networkx Digraph object describing nodes and edges
        """
        self.graph = graph
        self.paths = self.get_bitshift_paths()

    @property
    def node_ids(self):
        return set(self.graph.nodes)

    @staticmethod
    def load_from_links(links, task_name_mappings):
        """Converts a Links Dictionary into a Links object
        
        Args:
            links (list): List of Link descriptions matching app_schemas.LinkSchema
            task_name_mappings (dict): Mapping from node ID to task name
        
        Raises:
            DagHandlerValidationError: Raised if the resulting graph is not a valid DAG
        
        Returns:
            Links: Instantiated DAG object
        """

        graph = DiGraph()
        for link_dict in links:
            try:
                graph.add_edge(
                    task_name_mappings[link_dict["from_node"]["nodeId"]],
                    task_name_mappings[link_dict["to_node"]["nodeId"]],
                )
            except KeyError as e:
                raise DagHandlerValidationError(
                    f"Unable to find node(s) [{e}] referenced by links"
                )

        if not is_directed_acyclic_graph(graph):
            raise DagHandlerValidationError("Links do not form a valid DAG")

        return Links(graph)

    @staticmethod
    def graph_to_efficient_representation(graph):
        """Greedy algorithm to find an efficient representation of a graph.

        A Graph that looks like: 

            1 -> 2 -> 4 -> 5-> -->6
              -> 3 -------------|
                 \-> 7

        Will go to:
        [1, 2, 4, 5, 6]
        [1, 3, 7]
        [3, 6]
        """
        G = deepcopy(graph)
        paths = []
        while len(G.edges()) > 0:
            longest_path = dag_longest_path(G)
            paths.append(longest_path)
            edges = list(zip(longest_path[:-1], longest_path[1:]))

            for edge in edges:
                G.remove_edge(*edge)
        return paths

    def get_bitshift_paths(self):
        paths = self.graph_to_efficient_representation(self.graph)
        return [" >> ".join(path) for path in paths if len(path) > 1]


class DagHandler(_ParamHandler):

    docstring_parser = ClassParser()
    env = Environment(
        loader=PackageLoader("windmill", "templates"),
        autoescape=select_autoescape(enabled_extensions=["j2"]),
    )

    def __init__(self, dag_params: dict, tasks: List[TaskHandler], links: Links):
        """Handler to manage conversions between WMLs and Python DAGs. Should be 
        generated using DagHandler.load_from_wml
        
        Args:
            dag_params (dict): Dictionary of DAG parameters
            tasks (List[TaskHandler]): List of tasks metadata
            links (Links): Object describing links between tasks
        
        Raises:
            DagHandlerValidationError: Raised if WML can't be converted into a valid DAG instance
        """
        self.params = dag_params
        self.tasks = tasks
        self.links = links

    @property
    def snake_name(self):
        """snakecase version of task_params["dag_id"]
        
        Returns:
            [str]: dag_name
        """
        return underscore(self.params["dag_id"]["value"])

    @classmethod
    def marshall_dag_docstring(cls):
        return DagSchema().dump(cls.docstring_parser.parse_class(DAG)[1])

    @classmethod
    def load_from_wml(cls, wml_dict):
        """Returns a validated DagHandler from a Wml Dict
        
        Args:
            wml_dict (dict): Full WML file converted to a Dict
        
        Raises:
            DagHandlerValidationError: If instance params are invalid

        Returns:
            [DagHandler]: The converted DagHandler object
        """
        dag_params = cls.parameter_list_to_dict(wml_dict["dag"]["parameters"])
        if not dag_params.get("dag_id"):
            raise DagHandlerValidationError(f"DAG is missing required parameter dag_id")

        tasks = [
            TaskHandler.load_from_node(node) for node in wml_dict["nodes"].values()
        ]
        task_name_mappings = {t.node_id: t.snake_name for t in tasks}
        links = Links.load_from_links(wml_dict["links"].values(), task_name_mappings)

        return DagHandler(dag_params, tasks, links)

    @property
    def callable_tasks(self):
        return [task for task in self.tasks if task.python_callable]

    def compile_to_python(self):
        """Compiles the Dag Instance into Python code 
        
        Returns:
            [str]: The formatted DAG 
        """
        template = self.env.get_template("dag.j2")
        res = template.render(dag=self)
        return black.format_str(res, line_length=80)



import argparse
import logging
import os
import subprocess
from copy import deepcopy

from decli import cli

from ..config.project_config import ProjectConfig
from ..config.run_config import RunConfig
from ..http.api.endpoints import app
from ..tasks.init import CreateProject
from ..tasks.run import StartWebserver


class Cli:
    @staticmethod
    def get_cli_spec():
        return {
            "prog": "windmill",
            "description": "Drag'N'Drop web app to build and manage Airflow DAGs",
            "subcommands": {
                "title": "positional arguments",
                "description": "Run 'windmill <arg> --help' for further details",
                "commands": [
                    {
                        "name": "init",
                        "help": "Creates a new windmill project",
                        "func": Cli.init,
                        "arguments": ProjectConfig.to_cli_args(),
                    },
                    {
                        "name": "run",
                        "help": "Start Windmill server from a project folder",
                        "func": Cli.run_server,
                        "arguments": RunConfig.to_cli_args(),
                    },
                ],
            },
        }

    @classmethod
    def init(cls, *args, **kwargs):
        try:
            project = ProjectConfig.load(*args, **kwargs)
            CreateProject(project)
        except Exception as e:
            logging.error(f"Unable to create project ({e}) - aborting")
            return e

    @classmethod
    def run_server(cls, *args, **kwargs):
        try:
            run_config = RunConfig.load(*args, **kwargs)
            StartWebserver(run_config)
        except Exception as e:
            logging.error(f"Unable to start webserver ({e}) - aborting")

    @staticmethod
    def run_cli():
        return run_parser(get_parser(Cli.get_cli_spec()))


class DevCli:
    @staticmethod
    def get_cli_spec():
        return {
            "prog": "windmill",
            "description": "Dev commands for working on Windmill",
            "subcommands": {
                "title": "positional arguments",
                "description": "Run 'windmill <arg> --help' for further details",
                "commands": [
                    {
                        "name": "start-backend",
                        "help": "Starts the backend flask server with CORS enabled",
                        "func": DevCli.start_backend,
                        "arguments": RunConfig.to_cli_args(),
                    },
                    {
                        "name": "start-frontend",
                        "help": "Starts the frontend react server using npm build",
                        "func": DevCli.start_frontend,
                        "arguments": [],
                    },
                ],
            },
        }

    @staticmethod
    def start_backend(*args, **kwargs):
        try:
            wd = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
            print("Deleting existing windmill dev project")
            p = subprocess.Popen(
                ["rm", "-rf", ".windmill-temp-project/"], cwd=wd, stdout=subprocess.PIPE
            )
            p.communicate()

            print("Creating new project")
            p = subprocess.Popen(
                ["windmill", "init", "--name", ".windmill-temp-project"],
                cwd=wd,
                stdout=subprocess.PIPE,
            )
            p.communicate()

            print("Starting dev backend")
            os.chdir(
                os.path.abspath(
                    os.path.join(
                        os.path.dirname(__file__), "..", "..", ".windmill-temp-project/"
                    )
                )
            )
            run_config = RunConfig.load(run_dev_server=True, *args, **kwargs)
            StartWebserver(run_config)
        except Exception as e:
            logging.error(f"Unable to start webserver ({e}) - aborting")

    @staticmethod
    def start_frontend(**kwargs):
        wd = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "http", "app")
        )
        with subprocess.Popen(["npm", "start"], cwd=wd, stdout=subprocess.PIPE):
            print("Running frontend on http://localhost:1234")

    @staticmethod
    def run_cli():
        return run_parser(get_parser(DevCli.get_cli_spec()))


def get_parser(cli_spec) -> argparse.ArgumentParser:
    cli_spec["formatter_class"] = argparse.ArgumentDefaultsHelpFormatter
    return cli(cli_spec)


def run_parser(parser):
    args = parser.parse_args()

    try:
        args.func(**vars(args))
    except AttributeError:
        print(f"Error parsing args `{vars(args) or 'None'}`")
        parser.print_help()



import inspect
import logging
import pkgutil
from importlib import import_module

from airflow import operators

from .operator_handler import OperatorHandler
from ...exceptions import OperatorMarshallError


class OperatorIndex:
    def __init__(self, custom_operators=""):
        """Stateful object to index built-in and custom airflow
        operators

        Args:
            custom_operators (str, optional): Path to directory containing 
                custom operators. Defaults to "".
        """

        self.custom_operators = custom_operators
        self.operator_list = self.get_operators()

    def marshall_operator_list(self):
        """Return a JSON marshalled list of Operators as per OperatorHandler schema
        
        Returns:
            List[Dict]: List of OperatorHandler Dict - see `schemas.app_schemas.OperatorSchema`
        """
        handlers = []
        for operator in self.operator_list:
            try:
                handlers.append(OperatorHandler.from_operator(operator))
            except OperatorMarshallError as e:
                logging.exception(f"Unable to parse operator {operator.__name__}: {e}")

        return [h.dump() for h in handlers]

    def get_operators(self):
        """Get all default and custom operators
        
        Returns:
            List[Operator]: List of classes that inherit from BaseOperator
        """
        return self.get_default_operators()

    @staticmethod
    def get_default_operators():
        """Scrapes operators module for all classes that inherit from the BaseOperator
        class
        
        Returns:
            List[Operator]: List of Operator classes
        """
        ops = set()
        for _, modname, _ in pkgutil.iter_modules(operators.__path__):
            try:
                mod = import_module(f"airflow.operators.{modname}")
            except (ModuleNotFoundError, SyntaxError) as e:
                # NOTE Some of the HDFS libraries in Airflow don't support Python 3
                logging.info(f"Unable to import operator from {modname}: {e}")

            ops = ops.union(
                {
                    v
                    for v in mod.__dict__.values()
                    if inspect.isclass(v) and issubclass(v, operators.BaseOperator)
                }
            )

        return list(ops)



from ..config.run_config import RunConfig
from ..http.api.endpoints import build_app


class StartWebserver:
    def __init__(self, conf: RunConfig):
        app = build_app(conf.project_conf, dev_server=conf.run_dev_server)

        app.run(host=conf.hostname, port=conf.port)



__version__ = "0.0.2"

try:
    import airflow

    from . import cli
    from . import http
    from . import models
    from . import utils

except ImportError as e:
    print(
        "Airflow must be installed for Windmill to run. To install windmill with the tested version of airflow run:\n   pip install airflow-windmill[airflow]"
    )



import re

from marshmallow import Schema, fields


VALID_PARAMETER_TYPES = (
    "str",
    "dict",
    "list",
    "bool",
    "int",
    "float",
    "datetime.timedelta",
    "datetime.datetime",
    "callable",
)

PARAMETER_MAPPINGS = {
    "mapping": "dict",
    "python_callable": "callable",
    "python callable": "callable",
    "lambda": "callable",
    "function": "callable",
}


def validate_parameter_type(typ):
    return typ in VALID_PARAMETER_TYPES


class OperatorParameterSchema(Schema):
    id = fields.Str(required=True)
    type = fields.Str(validator=validate_parameter_type, required=True)
    value = fields.Str()
    default = fields.Str(allow_none=True)
    description = fields.Str()
    required = fields.Bool(default=False)
    inheritedFrom = fields.Str(allow_none=True)


class OperatorPropertiesSchema(Schema):
    name = fields.Str()
    module = fields.Str()
    description = fields.Str()
    parameters = fields.List(fields.Nested(OperatorParameterSchema()), required=True)


class OperatorSchema(Schema):
    type = fields.Str()
    properties = fields.Nested(OperatorPropertiesSchema())


class DagSchema(Schema):
    name = fields.Str(required=False)
    description = fields.Str()
    parameters = fields.List(fields.Nested(OperatorParameterSchema()), required=True)


class PositionSchema(Schema):
    x = fields.Int(required=True)
    y = fields.Int(required=True)


class PortSchema(Schema):
    id = fields.Str(required=True)
    type = fields.Str(required=True, validator=lambda typ: typ in ("top", "bottom"))
    properties = fields.Dict(keys=fields.Str(), values=fields.Str())
    position = fields.Nested(PositionSchema, required=True)


class SizeSchema(Schema):
    width = fields.Int()
    height = fields.Int()


class NodeSchema(Schema):
    id = fields.Str(required=True)
    position = fields.Nested(PositionSchema, required=True)
    orientation = fields.Int(required=True)
    type = fields.Str(required=True)
    ports = fields.Dict(keys=fields.Str(), values=fields.Nested(PortSchema))
    properties = fields.Nested(OperatorPropertiesSchema())
    size = fields.Nested(SizeSchema)


class LinkSpecSchema(Schema):
    nodeId = fields.Str(required=True)
    portId = fields.Str(required=True)


class LinkSchema(Schema):
    id = fields.Str(required=True)
    from_node = fields.Nested(LinkSpecSchema, data_key="from", required=True)
    to_node = fields.Nested(LinkSpecSchema, data_key="to", required=True)


class MinimalWmlSchema(Schema):
    # This is the minimal amount of data required to create an Airflow DAG
    filename = fields.Str(required=True)
    dag = fields.Nested(DagSchema, required=True)
    nodes = fields.Dict(
        keys=fields.Str(), values=fields.Nested(NodeSchema), required=True
    )
    links = fields.Dict(
        keys=fields.Str(), values=fields.Nested(LinkSchema), required=True
    )



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['windmill',
 'windmill.cli',
 'windmill.config',
 'windmill.http',
 'windmill.http.api',
 'windmill.models',
 'windmill.models.dags',
 'windmill.models.operators',
 'windmill.models.schemas',
 'windmill.tasks',
 'windmill.utils']

package_data = \
{'': ['*'],
 'windmill': ['templates/*'],
 'windmill.http': ['app/*',
                   'app/dist/*',
                   'app/public/*',
                   'app/src/*',
                   'app/src/components/*',
                   'app/src/components/Airflow/*',
                   'app/src/components/Navbar/*',
                   'app/src/components/Page/*',
                   'app/src/components/Sidebar/*',
                   'app/src/misc/*']}

install_requires = \
['black>=18.3-alpha.0,<19.0',
 'decli>=0.5.1,<0.6.0',
 'docstring-parser>=0.3.0,<0.4.0',
 'flask-cors>=3.0.8,<4.0.0',
 'inflection>=0.3.1,<0.4.0',
 'jinja2>=2.10.1,<3.0.0',
 'marshmallow>=3.0,<4.0',
 'networkx>=2.3,<3.0',
 'pyyaml>=5.1.2,<6.0.0']

extras_require = \
{'airflow': ['apache-airflow==1.10.4']}

entry_points = \
{'console_scripts': ['windmill = windmill:cli.cli.Cli.run_cli',
                     'windmill-dev = windmill:cli.cli.DevCli.run_cli']}

setup_kwargs = {
    'name': 'airflow-windmill',
    'version': '0.0.2',
    'description': "Drag'N'Drop Web Frontend for Building and Managing Airflow DAGs",
    'long_description': "# Windmill\n\n[![Build Status](https://travis-ci.org/mayansalama/windmill.svg?branch=master)](https://travis-ci.org/mayansalama/windmill)\n\nDrag'n'drop web app to manage and create Airflow DAGs. The goal is to\nhave a Web UI that can generate YML Dag Definitions, integrating with\ncustom operators and potentially existing DAGs. YML DAGs can then be\nsynced to a remote repo\n\n- Front end is built using React on Typescript\n- Back end is built using Flask on Python 3.6+\n\n## Getting Started\n\n1. Install with `pip install airflow-windmill`\n   1. Airflow is expected to be installed on the system. This allows Windmill to run with arbitrary versions of Airflow\n   2. Otherwise it can be packaged with windmill using `pip install airflow-windmill[airflow]`. The version is defined in `pyproject.toml`\n2. Run `windmill init` to create a local Windmill project\n3. `cd windmill-project`\n4. Run `windmill run` from this folder to run the app locally\n\n## MVP Required Features\n\n### Front-End Features\n\n- [x] Dynamic Operators\n- [x] Menu Dropdowns\n- [x] Load Operators from App\n- [x] Format operator display into classes\n- [x] Search functionality for operators\n- [x] Basic operator level properties\n- [x] Implement DAG level properties\n- [x] New DAG Functionality\n- [x] Parameter Tooltips\n- [x] Render arbitrary viewport windows for New/Save/Load etc\n- [x] Overwrite/Save prompt on New\n- [x] DAG renaming and save functionality\n- [x] Open dag from menu\n- [x] Make save/load more efficient by removing non-essential values\n- [x] Switch nav menu to icons \n- [x] Add convert DAG call\n- [ ] Add hotkeys to menu functions\n- [ ] Make input/output nodes more clear\n- [ ] Check if file already exists on rename\n- [ ] Prompt save if there are nodes on open\n- [x] Fix loss of state on refresh bug\n- [ ] Put File details in File Browse\n- [ ] Make Flask Backend URI configurable\n- [ ] Add a last saved time to NavBar\n- [ ] Add error handling to backend calls\n- [ ] Add tests\n- [ ] Get task descriptions from Operator list\n\n### Back-End Features\n\n- [x] Generate Operator Lists\n- [x] CLI to start Web and Front End\n- [x] Generate DAG Spec\n- [x] CLI to create new windmill project\n- [x] CLI to start windmill from a windmill project\n- [x] Implement windmill-dev start\n- [x] Save/Load Windmill Files functionality\n- [x] Get default values\n- [x] Pull parameters from parent classes\n- [x] Move airflow dependency as extra\n- [x] Convert WML into Python DAG\n- [x] API Endpoint to trigger WML -> DAG\n- [ ] Edge cases for WML -> DAG\n- [ ] Get WML owner and last-modified details during wml list\n- [ ] Allow custom operators\n- [ ] Add defaults to CLI --help commands\n- [ ] Strategy for Python Opjects (e.g. callables) - allow either a import ref or an inline statement\n- [ ] Backport existing Python DAGs to WMLs\n- [ ] Allow DAG updates to propogate to WMLs\n- [ ] Add tests for different airflow version\n\n### Other features\n\n- [ ] Validate on backend or front end or both?\n- [ ] Doco\n- [ ] Add permission restrictions for valid tags \n\n## Dev User Guide\n\nTo run as a dev:\n\n1. Clone from git\n2. Run `poetry install -E airflow`\n3. Run `windmill-dev start-backend`\n4. Run `windmill-dev start-frontend`\n\n## Future Usage Patterns\n\n- Auto-sync for windmill project to git\n\n## Deployment\n\nDeployment to PyPi is managed using Travis and should be done in the following steps:\n\n1. Run `poetry version {patch|minor|major}`\n2. Increment the version number in `windmill/__init__.py`\n3. Commit and merge code into the master branch\n4. Ensure that the travis build is green\n5. Create a git tag for the new build\n6. Push the tag to origin\n",
    'author': 'mayansalama',
    'author_email': 'micsalama@gmail.com',
    'url': 'https://github.com/mayansalama/windmill',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import logging
import re

from airflow.operators import BaseOperator
from docstring_parser import parse
from marshmallow import schema

from ..schemas.app_schemas import (
    OperatorSchema,
    validate_parameter_type,
    VALID_PARAMETER_TYPES,
)
from ...utils import class_parser
from ... import exceptions


class OperatorHandler:
    """Handler to translate between Schemas, Operator Objects and
    dictionaries
    """

    schema = OperatorSchema()
    docstring_parser = class_parser.ClassParser()

    def __init__(self, type: str, properties: dict):
        """Should initialise using the from_marsh or from_operator method

        Args:
            type (str):  Operator Name
            properties (dict): Operator Properties
        """
        self.type = type
        self.properties = properties

    def dump(self):
        """Serialise to dict

        Returns:
            Dict: Marshalled result
        """
        return self.schema.dump(self)

    @classmethod
    def from_operator(cls, operator: BaseOperator.__class__):
        """Instantiate a handler from an Airflow Operator

        Arguments:
            operator {BaseOperator} -- Airflow Operator Class
        """
        try:
            return OperatorHandler(
                *cls.docstring_parser.parse_class(operator, inherit_until=BaseOperator)
            )
        except exceptions.DocstringParseError as e:
            raise exceptions.OperatorMarshallError(
                f"Unable to parse class {operator.__name__}"
            ) from e

    @staticmethod
    def from_dict(d: dict):
        """[summary]
        
        Args:
            d (dict): Unparsed dict
        
        Returns:
            OperatorHandler: Parsed operator handler
        """
        return OperatorHandler.from_marsh(OperatorHandler.schema.load(d))

    @staticmethod
    def from_marsh(res: dict):
        """Instantiate a handler from an Unmarshal Result

        Args:
            data (dict): Result from Marshmallow.load
        
        Returns:
            OperatorHandler: Parsed operator handler
        """
        return OperatorHandler(res["type"], res["properties"])
